#ifndef TEMPLATEDATABASE_H
#define TEMPLATEDATABASE_H

#include <iostream>
#include <fstream>
#include <vector>
using namespace std;

class TemplateDatabase
{
public:

	TemplateDatabase();
	~TemplateDatabase();
	void setDatabase( unsigned long int numInvFile, unsigned int numImages);
	int allocatePointerArray();
	int allocateInvertedFile(unsigned int invFileIndex);
	int addImageToInvertedFile(unsigned int imgID, unsigned int invFileIndex);
	int incramentmi(unsigned int imgID, unsigned long int invFileIndex);
	void incramentNumberOfImages();
	unsigned int getmi(unsigned int imgID, unsigned long int invFileIndex);
	unsigned int getLastImgID(unsigned long int invFileIndex);
	unsigned int getNumberOfImages();
	unsigned int getNi(unsigned long int invFileIndex);
	double getScore(unsigned int imgID);
	void updateQueryTree(unsigned int invFileIndex, unsigned int descIndex);
	unsigned int searchQueryTree(unsigned long int invFileIndex);
	void deleteQuery();
	void updateForwardFileTree(unsigned int imgID, unsigned int invFileIndex);
	unsigned int getMatch();
	void saveDatabase();
	void loadDatabase(char *fileName);
	void resetInvertedFile(unsigned long int invFileIndex);
	void resetForwardFileList();
	vector<int> getDescIndexList(unsigned int imgID);
	void getImgScoreList(vector<int>& idVector);
	void updateObjectIDS(unsigned int imgID, unsigned int nextImgID);
	vector<int> TemplateDatabase::getObjectIDS(unsigned int imgID);

	unsigned int N;

private:

	typedef struct DescIndexST
	{
		unsigned int index;
		struct DescIndexST *nextDescIndex;
	} *DescIndex;

	typedef struct DescIndexSetST
	{
		DescIndex indexList;
		struct DescIndexSetST *nextIndexList;
	} *DescIndexSet;

	typedef struct ObjectIDSST
	{
		unsigned int ID;
		struct ObjectIDSST *nextObjectIDS;
	} *ObjectIDS;

	typedef struct InvFileIndexLinkST
	{
		unsigned long int invFileIndex;
		unsigned int mi;
		struct InvFileIndexLinkST *leftInvFileIndexLink;
		struct InvFileIndexLinkST *rightInvFileIndexLink;
	} *InvFileIndexLink;

	typedef struct ImgNodeST
	{
		unsigned int imgID;
		InvFileIndexLink invFileIndexTree;
		ObjectIDS objectIDSList;
		ObjectIDS objectIDSListEnd;
		struct ImgNodeST *nextImgNode;
	} *ImgNode;

	typedef struct ScoreNodeST
	{
		unsigned int imgID;
		double score;
		DescIndexSet indexSet;
		DescIndexSet indexSetEnd;
		struct ScoreNodeST *leftScoreNode;
		struct ScoreNodeST *rightScoreNode;
	} *ScoreNode;

	typedef struct QueryNodeST
	{
		unsigned long int LeafNodeID;
		unsigned int ni;
		DescIndex descList;
		DescIndex descListEnd;
		struct QueryNodeST *leftQueryNode;
		struct QueryNodeST *rightQueryNode;
	} *QueryNode;

	typedef struct ImageLinkST
	{
		unsigned int ID;
		//unsigned int mi;
		struct ImageLinkST *nextImageLink;
	} *ImageLink;

	typedef struct InvertedFileST
	{
		ImageLink imgListHead;
		ImageLink *lastImageLink;
        unsigned int Ni;
	} *InvertedFile;

    InvertedFile *invFileArray;
	unsigned long int numInvFile;
	unsigned int N_ent;
	QueryNode queryTree;
	ImgNode forwardFileList;//List of forward files.
	ScoreNode scoreTree;
	bool init;
	double qNorm , dNorm;

	void updateQuery(QueryNode queryNode, unsigned long int invFileIndex, unsigned int descIndex);
	void deleteQueryTree(QueryNode queryNode);
	void deleteScoreTree(ScoreNode scoreNode);
	void updateForwardFile(ImgNode* imgNode, unsigned int imgID, unsigned long int invFileIndex);
	void updateScoreTree(unsigned int imgID, double score);
	void calcQNorm(QueryNode tempQueryNode);
	double getNodeWeight(unsigned long int invFileIndex);
	void calcDNorm(InvFileIndexLink tempInvFileIndexLink);
	void calcScore(ScoreNode tempScoreNode);
	void calcImgScore(ScoreNode tempScoreNode);
	void queryInvFileScore(InvFileIndexLink tempInvFileIndexLink, QueryNode tempQueryNode, ScoreNode tempScoreNode);
	void calcQueryScore(QueryNode tempQueryNode, InvFileIndexLink tempInvFileIndexLink, ScoreNode tempScoreNode);
	void saveInvertedFile(ofstream& myfile);
	void loadInvertedFile(FILE *myfile);
	int addNi(unsigned int invFileIndex, unsigned int addNi);
	void loadForwardFile(FILE *myfile);
	void saveForwardFile(ofstream& myfile);
	void saveInvFileIndexTree(InvFileIndexLink tempInvFileIndexLink, ofstream& myfile);
	void deleteInvFileIndexLink(InvFileIndexLink tree);
	void deleteImgNode(ImgNode tempImgNode);
	void deleteObjectIDS(ObjectIDS tempObjectIDS);
	void deleteImageLinkList(ImageLink tempImageLink);
	void getImgScoreIDS(ScoreNode tempScoreNode, vector<int>& idVector);
};

#endif