#ifdef WIN32
#define _CRT_SECURE_NO_DEPRECATE
#endif
#include <math.h>
#include <iostream>
#include <fstream>
#include <cstring>
#include "VocabularyTree.h"



VocabularyTree::VocabularyTree()
{
    dim = 0;
    levels = 0;
    clusters = 0;
    total_nodes = 0;
    init = false;
    tree = NULL;
}

VocabularyTree::VocabularyTree(unsigned char dim, unsigned char levels, unsigned char clusters)
{
    this->dim = dim; /*Sets dim to the number of dimension passed*/
    this->levels = levels; /*Sets levels to the number of levels passed*/
    this->clusters = clusters; /*Sets clusters to the number of clusters passed*/
    total_nodes = 0;
    init = false;
    tree = NULL;
}

VocabularyTree::~VocabularyTree()
{
    delete tree;
}

void VocabularyTree::setDimension(unsigned char dim)
{
    this->dim = dim; /*Sets dim to the number of dimension passed*/
}

void VocabularyTree::setLevel(unsigned char levels)
{
    this->levels = levels; /*Sets levels to the number of levels passed*/
}

void VocabularyTree::setNumberOfClusters(unsigned char clusters)
{
    this->clusters = clusters; /*Sets clusters to the number of clusters passed*/
}

void VocabularyTree::initTree(unsigned char dim, unsigned char levels, unsigned char clusters)
{
    this->dim = dim; /*Sets dim to the number of dimension passed*/
    this->levels = levels; /*Sets levels to the number of levels passed*/
    this->clusters = clusters; /*Sets clusters to the number of clusters passed*/
}

int VocabularyTree::allocateTree()
{

    if(levels == 0 || dim == 0 || clusters == 0)
        return 0; //return zero if level or dim or clusters are set to zero

    unsigned int nodeCount;
    try 
    {
        tree = (VocabTree) new VocabTree;//allocate the tree
        tree->level = (TreeLevel*)(new TreeLevelSt[levels]); //allocate the number of levels

        for(int i = 0; i < levels; i++) //i is the current level
        {
            tree->level[i] = (TreeLevel)(new TreeLevelSt);//allocate the levels in the tree

            nodeCount = (unsigned int)pow( (double)clusters, (double)(i+1) ); //nodeCount is the
            //number of nodes for the current level i.

            tree->level[i]->nodes = (Node*)(new NodeSt[ nodeCount ]); //allocate the nodes for the current level i


            for(unsigned int nodeIndex = 0; nodeIndex < nodeCount; nodeIndex++)
            {
				tree->level[i]->nodes[nodeIndex] = NULL;

               /* tree->level[i]->nodes[nodeIndex] = (Node) malloc(sizeof(struct NodeSt));
                tree->level[i]->nodes[nodeIndex]->descVector = (unsigned char *) malloc(dim);*/
            }
        }
    }
    catch( char * str )
    {
        cerr << "Exception raised: " << str << '\n';
        init = false;
        return 0;
    }

    init = true;
    return 1;
}

unsigned char VocabularyTree::getDescVectorElement(unsigned char levelIndex, unsigned int nodeIndex, unsigned char elementIndex)
{
    return (tree->level[levelIndex-1]->nodes[nodeIndex-1]->descVector[elementIndex-1]);
}

bool VocabularyTree::nodeExist(unsigned char levelIndex, unsigned int nodeIndex)
{
    if(tree->level[levelIndex-1]->nodes[nodeIndex-1] != 0)
		return true;

	return false;
}

void VocabularyTree::allocateNode(unsigned char levelIndex, unsigned int nodeIndex)
{
	if(tree->level[levelIndex-1]->nodes[nodeIndex-1] == 0){

		tree->level[levelIndex-1]->nodes[nodeIndex-1] = (Node) malloc(sizeof(struct NodeSt));
		tree->level[levelIndex-1]->nodes[nodeIndex-1]->descVector = (unsigned char *) malloc(dim);
	}
	return;
}

unsigned char* VocabularyTree::getDescVector(unsigned char levelIndex, unsigned int nodeIndex)
{
    return (tree->level[levelIndex-1]->nodes[nodeIndex-1]->descVector);
}

void VocabularyTree::setDescVectorElement(unsigned char levelIndex, unsigned int nodeIndex, unsigned char elementIndex, unsigned char element)
{
    tree->level[levelIndex-1]->nodes[nodeIndex-1]->descVector[elementIndex-1] = element;
}

void VocabularyTree::setNodeToNull(unsigned char levelIndex, unsigned int nodeIndex)
{
    delete tree->level[levelIndex-1]->nodes[nodeIndex-1];
	tree->level[levelIndex-1]->nodes[nodeIndex-1] = 0;
}

void VocabularyTree::setDescVector(unsigned char levelIndex, unsigned int nodeIndex, unsigned char* descVector)
{
    unsigned char currentLevel = levelIndex-1;
    unsigned int currentNode = nodeIndex-1;

    //set each descriptor vector element
    //for(int i = 0; i < dim; i++)
    //{
    //	tree->level[currentLevel]->nodes[currentNode]->descVector[i] = descVector[i];
    //}

    //set the descriptor vector pointer
    tree->level[currentLevel]->nodes[currentNode]->descVector = descVector;
}

unsigned int VocabularyTree::getParentArrayIndex(unsigned char levelIndex, unsigned int nodeIndex)
{
	if((levelIndex-1) == 0)//check if we are in the first level
		return -1;//return -1 if we are in the first level

	unsigned int nodeCount = (unsigned int)pow( (double)clusters , (double)levelIndex );

	if((nodeIndex-1) >= nodeCount || (nodeIndex-1) < 0
		|| (levelIndex-1) >= levels || (levelIndex-1) < 0)//check if the index is out of bound
		return -2; //return -2 if index is out of bound

	return	(unsigned int)floor((double)((nodeIndex-1)/clusters));
}

unsigned int VocabularyTree::getChildArrayIndex(unsigned char levelIndex, unsigned int nodeIndex, unsigned char child)
{
	if(levelIndex == levels)//check if we are in the last level
		return -1;//return -1 if we are in the last level

	unsigned int nodeCount = (unsigned int)pow( (double)clusters , (double)levelIndex );

	if((nodeIndex-1) >= nodeCount || (nodeIndex-1) < 0
		|| (child - 1) >= clusters || (levelIndex-1) >= levels || (levelIndex-1) < 0)//check if the index is out of bound
		return -2; //return -2 if index is out of bound

	return	(unsigned int)(((nodeIndex-1)*clusters) + (child - 1));
}

//void VocabularyTree::writeTreeFile( char *file_loc )  //Dump Tree File
//{
//
//	using namespace std;
//
//	double index = 1,
//		   previous_lvl_index = 1;
//
//	int current_lvl = 1,
//		descriptor_cnt = 0,
//		relative_index=1,
//		new_levels = levels + 1;
//
//	//Calculates Total # of nodes and stores it in class
//	calcNumNodes();
//
//	//Create File
//	ofstream myfile;
//	myfile.open( file_loc );
//
//	//Write first line of file, info about tree
//	myfile << (int)levels << " " << (int)clusters << " " << total_nodes << " " << (int)dim << endl;
//		
//	int l = 1;
//	for( int i = 1; i <= this->levels; i++ )
//	{
//		for( double j = 1; j <= pow( (double)this->clusters , (double)(i) ); j++ )
//		{
//
//			myfile << i-1 << " " << j-1;
//			if(this->nodeExist(i,j)){
//				for( int k = 1; k <= this->dim; k++ )
//				{
//					myfile << " " << (int)this->getDescVectorElement( i , j , k );
//					l++;
//				}
//			}
//			else
//				myfile << " " << int(-1);
//
//			myfile << endl;
//		}
//	}
//	myfile.close();
//}
//
//void VocabularyTree::retrieveTreeIntoMem( char *file_loc )
//{
//
//	using namespace std;
//
//	unsigned int levels,clusters,dim;
//	float total_nodes;
//
//	FILE * file;
//	file = fopen( file_loc , "r" );
//
//	//Sets levels, clusters, total_nodes and dimensions
//	fscanf( file , "%d %d %f %d\n" , &levels , &clusters , &total_nodes , &dim );
//	this->setLevel( levels );
//	this->setNumberOfClusters( clusters );
//	this->setDimension( dim );
//	this->calcNumNodes();
//	this->allocateTree();
//
//	unsigned int nodeLevel=0,
//				 nodeIndex=0;
//
//	int dim_temp=0;	
//
//	//Loop through each line (node) after root node
//	for( int i = 1; i <= this->levels; i++ )
//	{
//		for( double j = 1; j <= pow( (double)this->clusters , (double)(i) ); j++ )
//		{
//			unsigned int nodeLevel=0,
//				nodeIndex=0,
//				dim_temp=0;
//
//			fscanf( file , "%d %d " , &nodeLevel , &nodeIndex );
//			for( int k = 1; k <= this->dim; k++ )
//			{
//				fscanf( file , "%d " , &dim_temp );
//				if(dim_temp != -1){
//					this->allocateNode( i, j );
//					this->setDescVectorElement( i , j , k , dim_temp ); 
//				}
//				else
//					break;
//			}
//		}
//	}
//	fclose( file );
//}

int VocabularyTree::calcNumNodes()
{
	total_nodes = 0;
	if( clusters == 0 || levels == 0 )
	{
		return 0;
	}    

	for( int i = 0; i <= (int)levels; i++ )
	{
		total_nodes += pow( (double)clusters , i );
	}
    total_nodes--;  //Comment this line to count root node in node count

	return 1;
}

void VocabularyTree::binaryModeWriteTreeFile( char *file_loc )  //Dump Tree File
{

	FILE * myfile;
	myfile = fopen ( file_loc , "wb" );	

	fputc ( (int) levels , myfile );
	fputc ( (int) clusters , myfile );
	fputc ( (int) dim , myfile );

	for( unsigned char i = 1; i <= this->levels; i++ )
	{
		for(unsigned int j = 1; j <= pow( (double)this->clusters , (double)(i) ); j++ )
		{

			if(this->nodeExist(i,j)){

				unsigned char tempUChar = i-1;
				fputc ( (int) tempUChar , myfile );

				for(int tempI = 6; tempI >= 0; tempI-=2){
					tempUChar = (j-1) >> 4*tempI;
					fputc ( (int) tempUChar , myfile );
				}

				fwrite(this->getDescVector(i,j), 1, dim, myfile);
			}
		}
	}
	fclose(myfile);
}

void VocabularyTree::textModeWriteTreeFile( char *file_loc )  //Dump Tree File
{

	using namespace std;

	double index = 1,
		   previous_lvl_index = 1;

	int current_lvl = 1,
		descriptor_cnt = 0,
		relative_index=1,
		new_levels = levels + 1;

	//Calculates Total # of nodes and stores it in class
	calcNumNodes();

	//Create File
	ofstream myfile;
	myfile.open( file_loc );

	//Write first line of file, info about tree
	myfile << (int)levels << " " << (int)clusters << " " << total_nodes << " " << (int)dim << endl;
		
	int l = 1;
	for( int i = 1; i <= this->levels; i++ )
	{
		for( double j = 1; j <= pow( (double)this->clusters , (double)(i) ); j++ )
		{

			myfile << i-1 << " " << j-1;
			if(this->nodeExist(i,j)){
				for( int k = 1; k <= this->dim; k++ )
				{
					myfile << " " << (int)this->getDescVectorElement( i , j , k );
					l++;
				}
			}
			else
				myfile << " " << int(-1);

			myfile << endl;
		}
	}
	myfile.close();
}

void VocabularyTree::binaryModeRetrieveTreeIntoMem( char *file_loc )
{
	FILE * file;
	file = fopen( file_loc , "rb" );

	//Sets levels, clusters, total_nodes and dimensions
	fscanf( file , "%c%c%c" , &levels , &clusters, &dim );

	this->calcNumNodes();
	this->allocateTree();

	//Loop through each line (node) after root node
	while(1){

		int nodeLevel = 0, dim_temp = 0;
		unsigned int nodeIndex = 0;

		nodeLevel = fgetc(file);

		for(int i = 6; i >= 0; i-=2){
			unsigned int temp = (unsigned int)fgetc(file);
			nodeIndex += temp << 4*i;
		}

		// If the end of file was reached, break out of the while loop
		if(feof(file))
			break;

		this->allocateNode( nodeLevel+1, nodeIndex+1 );

		for( unsigned char k = 1; k <= this->dim; k++ )
		{
			dim_temp = fgetc(file);
			this->setDescVectorElement( nodeLevel+1 , nodeIndex+1 , k , dim_temp );
		}

	}

	fclose( file );
}

void VocabularyTree::textModeRetrieveTreeIntoMem( char *file_loc )
{

	using namespace std;

	unsigned int levels,
		clusters,
		total_nodes,
		dim;	

	FILE * file;
	file = fopen( file_loc , "r" );

	//Sets levels, clusters, total_nodes and dimensions
	fscanf( file , "%d %d %d %d\n" , &levels , &clusters , &total_nodes , &dim );
	this->setLevel( levels );
	this->setNumberOfClusters( clusters );
	this->setDimension( dim );
	this->calcNumNodes();
	this->allocateTree();

	unsigned int nodeLevel=0,
				 nodeIndex=0;

	int dim_temp=0;	

	//Loop through each line (node) after root node
	for( int i = 1; i <= this->levels; i++ )
	{
		for( double j = 1; j <= pow( (double)this->clusters , (double)(i) ); j++ )
		{
			unsigned int nodeLevel=0,
				nodeIndex=0,
				dim_temp=0;

			fscanf( file , "%d %d " , &nodeLevel , &nodeIndex );
			for( int k = 1; k <= this->dim; k++ )
			{
				fscanf( file , "%d " , &dim_temp );
				if(dim_temp != -1){
					this->allocateNode( i, j );
					this->setDescVectorElement( i , j , k , dim_temp ); 
				}
				else
					break;
			}
		}
	}
	fclose( file );
}